package l10_concurrent;

import utilities.RandomGenerators;
import utilities.ThreadHelperMethods;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/*
	A BlockingQueue is typically used to have on thread produce objects, which
	another thread consumes.
	The producing thread will keep producing new objects and insert them into
	the queue, until the queue reaches some upper bound on what it can contain.
	It's limit, in other words. If the blocking queue reaches its upper limit,
	the producing thread is blocked while trying to insert the new object. It
	remains blocked until a consuming thread takes an object out of the queue.
	The consuming thread keeps taking objects out of the blocking queue, and
	processes them. If the consuming thread tries to take an object out of an
	empty queue, the consuming thread is blocked until a producing thread puts
	an object into the queue.
*/

public class BlockingQueuePC {
	public static void main(String[] args) {
		final int numberOfProducers = (args.length > 0) ? Integer.parseInt(args[0]) : 2;
		final int queueCapacity = (args.length > 1) ? Integer.parseInt(args[1]) : 5;
		final int numberOfThreads = Runtime.getRuntime().availableProcessors();
		final BlockingQueue<Integer> queue = new ArrayBlockingQueue<>(queueCapacity);

		try (ExecutorService executorService = Executors.newFixedThreadPool(numberOfThreads)) {
			for (int i = 0; i < numberOfProducers; ++i) {
				executorService.submit(new Producer("Producer " + i, queue));
				executorService.submit(new Consumer("Consumer " + i, queue));
			}
		}
	}

// Producer class.
// It sleeps 1 - 2 seconds between each put() call.
// This will cause the Consumer to block, while waiting for
// objects in the queue. 
	record Producer(String name, BlockingQueue<Integer> queue) implements Runnable {
		@Override
		public void run() {
			for (int i = 0; i < 5; ++i) {
				int number = RandomGenerators.getIntThread(100);
				ThreadHelperMethods.safeSleep(1000, 2000);
				System.out.println(name + " putting " + number);
				try {
					queue.put(number);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}

// Consumer class.
// It takes out the objects from the queue and prints them to System.out.
	record Consumer(String name, BlockingQueue<Integer> queue) implements Runnable {
		@Override
		public void run() {
			for (int i = 0; i < 5; ++i) {
				ThreadHelperMethods.safeSleep(1500, 3000);
				try {
					System.out.println("\t" + name + " getting " + queue.take());
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}
}

