package l10_concurrent;

import utilities.RandomGenerators;
import utilities.ThreadHelperMethods;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;

//	ConcurrentLinkedQueue is an unbounded thread-safe queue based on linked
//	nodes. This queue orders elements as a FIFO.

//	A ConcurrentLinkedQueue is a good choice when many threads share access to a
//	common collection.

public class ConcurrentLinkedQueueExample {
	public static void main(String[] args) {
		int numberOfProducers = (args.length > 0) ? Integer.parseInt(args[0]) : 4;
		int numberOfThreads = Runtime.getRuntime().availableProcessors();

		ConcurrentLinkedQueue<Integer> queue = new ConcurrentLinkedQueue<>();
		try (ExecutorService executorService = Executors.newFixedThreadPool(numberOfThreads)) {
			for (int i = 0; i < numberOfProducers; i++) {
				executorService.submit(new Producer("Producer " + i, queue));
				executorService.submit(new Consumer("Consumer " + i, queue));
			}
		}
	}

/*
	BlockingQueue blocks the consumer or the producer when the queue is
	empty or full and the respective consumer/producer thread is put to sleep.
	But this blocking feature comes with a cost: every put or take operation is
	lock contended between the producers or consumers (if many), so in scenarios
	with many producers/consumers the operation might be slower.

	ConcurrentLinkedQueue is not using locks. But being a "wait free" data
	structure, ConcurrentLinkedQueue will not block when empty, meaning that the
	consumer will need to deal with the poll() returning null values by "busy
	waiting", for example, with the consumer thread eating up CPU.

	So which one is "better" depends on the number of consumer threads, on the
	rate they consume/produce, etc. A benchmark is needed for each scenario.
	One particular use case where the ConcurrentLinkedQueue is clearly better is
	when producers first produce something and finish their job by placing the
	work in the queue and only after the consumers start to consume, knowing
	that they will be done when queue is empty (here is no concurrency between
	producer-consumer but only between producer-producer and consumer-consumer)
	the producer puts strings on the queue
*/

	record Producer(String name, ConcurrentLinkedQueue<Integer> queue) implements Runnable {
		@Override
		public void run() {
			for (int i = 0; i < 5; i++) {
				ThreadHelperMethods.safeSleep(1000, 2000);
				int element = RandomGenerators.getIntThread(100);
				queue.add(element);
				System.out.println(name + " added: " + element);
			}
		}
	}

	record Consumer(String name, ConcurrentLinkedQueue<Integer> queue) implements Runnable {
		@Override
		public void run() {
			Integer number;
			for (int i = 0; i < 5; i++) {
				ThreadHelperMethods.safeSleep(1500, 3000);
				while ((number = queue.poll()) == null) ; // busy waiting while queue empty
				System.out.println("\t" + name + " removed: " + number);
			}
		}
	}
}

