package l10_concurrent;

import utilities.ThreadHelperMethods;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/*
	 A java.util.concurrent.CountDownLatch is a concurrency construct that
	 allows one or more threads to wait for a given set of operations to
	 complete.

	 A CountDownLatch is initialized with a given count. This count is
	 decremented by calls to the countDown() method. Threads waiting for this
	 count to reach zero can call one of the await() methods. Calling await()
	 blocks the thread until the count reaches zero.
*/

public class CountDownLatchExample {

	private static final int numberThreads = Runtime.getRuntime().availableProcessors();
	// Latch opens when all threads started - we want the threads to wait until all other threads are ready
	private static final CountDownLatch readyThreadCounter = new CountDownLatch(numberThreads);
	// Latch opens when all threads are done
	private static final CountDownLatch completedThreadCounter = new CountDownLatch(numberThreads);
	// Latch blocks threads until all threads started
	private static final CountDownLatch callingThreadBlocker = new CountDownLatch(1);

	public static void main(String[] args) {

		try (ExecutorService executorService = Executors.newFixedThreadPool(numberThreads)) {
			for (int i = 0; i < numberThreads; ++i) {
				executorService.submit(new CDLThread("Thread " + i));
			}

			// wait for the threads to start
			ThreadHelperMethods.awaitCountDownLatch(readyThreadCounter);
			System.out.println("All threads started");
			// unblock threads
			callingThreadBlocker.countDown();
		}

		// wait for the threads to finish
		ThreadHelperMethods.awaitCountDownLatch(completedThreadCounter);
		System.out.println("All threads completed");
	}

	record CDLThread(String name) implements Runnable {

		@Override
		public void run() {
			// thread started
			readyThreadCounter.countDown();
			// wait for remaining threads to start
			ThreadHelperMethods.awaitCountDownLatch(callingThreadBlocker);
			System.out.println("\t" + name + " starting");
			// get the job done
			ThreadHelperMethods.safeSleep(1000, 4000);
			System.out.println("\t\t" + name + " finished");
			// thread ready; decrement latch counter.
			completedThreadCounter.countDown(); // thread finished
		}
	}
}

