package l10_concurrent;

import utilities.RandomGenerators;
import utilities.ThreadHelperMethods;

import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class PCMutex {
	private static final int numberOfProducers = 2;
	private static final int numberOfLoops = 5;

	public static void main(String[] args) {
		Buffer sharedBuffer = new Buffer();
		int n = Runtime.getRuntime().availableProcessors();
		try (ExecutorService executorService = Executors.newFixedThreadPool(n)) {
			for (int i = 0; i < numberOfProducers; i++) {
				executorService.submit(new Producer(sharedBuffer));
				executorService.submit(new Consumer(sharedBuffer));
			}
		}
	}

/*
	Like threads communicate using wait(), notify() and notifyAll() methods of object
	when intrinsic lock is used to restrict access to a shared objects, Condition allows
	inter-threads communication when external locks are used. Condition defines mechanism
	to suspend a thread until another thread notifies it.

	Using condition, deadlock issue can be avoided with external locks.
*/

	static class Buffer {
		private final int bufferSize = 1; // size of the buffer array
		private final int[] buffer = new int[bufferSize]; // an array of Objects
		private final Lock lock = new ReentrantLock();
		private final Condition isEmpty = lock.newCondition();
		private final Condition isFull = lock.newCondition();
		private int current = 0; // number of items currently in the buffer

		private void put(int item) {
			lock.lock();
			try {
				while (current >= buffer.length) {
					isFull.await(); // InterruptedException
				}
				buffer[current++] = item;
				isEmpty.signal();
				System.out.format("Producer %3d put %3d%n", Thread.currentThread().threadId(), item);
			} catch (InterruptedException e) {
				e.printStackTrace();
			} finally {
				lock.unlock();
			}
		}

		private int get() {
			lock.lock();
			try {
				while (current <= 0) {
					isEmpty.await();
				}
				int item = buffer[--current];
				isFull.signal();
				System.out.format("\tConsumer %3d got %3d%n", Thread.currentThread().threadId(), item);
				return item;
			} catch (InterruptedException e) {
				e.printStackTrace();
				return 0;
			} finally {
				lock.unlock();
			}
		}
	}

	record Producer(Buffer buffer) implements Runnable {
		@Override
		public void run() {
			for (int loop = 0; loop < numberOfLoops; ++loop) {
				ThreadHelperMethods.safeSleep(1000);
				buffer.put(RandomGenerators.getIntThread(100));
			}
		}
	}

	record Consumer(Buffer buffer) implements Runnable {
		@Override
		public void run() {
			for (int loop = 0; loop < numberOfLoops; ++loop) {
				ThreadHelperMethods.safeSleep(1500);
				buffer.get();
			}
		}
	}
}

