package l10_concurrent;

import utilities.RandomGenerators;
import utilities.ThreadHelperMethods;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/*
	This class demonstrates how to use ReadWriteLock to add concurrency
	features to a non-threadsafe collection

	A ReadWriteLock maintains a pair of associated locks, one for read-only
	operations and one for writing. The read lock may be held simultaneously by
	multiple reader threads, so long as there are no writers. The write lock is
	exclusive.

	Policy issues:
	Determining whether to grant the read lock or the write lock, when both
	readers and writers are waiting, at the time that a writer releases the
	write lock. Writer preference is common, as writes are expected to be short
	and infrequent. Reader preference is less common as it can lead to lengthy
	delays for a write if the readers are frequent and long-lived as expected.
	Fair, or "in-order" implementations are also possible.
*/

public class ReadWriteLockTest {
	private static final List<Integer> list = new ArrayList<>();
	private static final ReadWriteLock rwLock = new ReentrantReadWriteLock();

	public static void main(String[] args) {
		final int numberOfReaders = 16;
		final int numberOfWriters = 8;
		int n = Runtime.getRuntime().availableProcessors();

		try (ExecutorService executorService = Executors.newFixedThreadPool(n)) {
			for (int i = 0; i < numberOfWriters; i++) {
				executorService.submit(new Writer());
			}
			for (int i = 0; i < numberOfReaders; i++) {
				executorService.submit(new Reader());
			}
		}
	}

	static class Writer implements Runnable {

		private void add(int number) {
			rwLock.writeLock().lock();
			try {
				ThreadHelperMethods.safeSleep(500, 2000);
				list.add(number);
			} finally {
				rwLock.writeLock().unlock();
			}
		}

		@Override
		public void run() {
			int number = RandomGenerators.getIntThread(100);
			add(number);
			System.out.println("Writer " + Thread.currentThread().threadId() + " -> put: " + number);
		}
	}

	static class Reader implements Runnable {

		private int get() {
			rwLock.readLock().lock();
			try {
				ThreadHelperMethods.safeSleep(500, 2000);
				// Throws: IllegalArgumentException - if bound is not positive (list empty)
				return list.get(RandomGenerators.getIntThread(list.size()));
			} catch(IllegalArgumentException e) {
				System.out.println("List empty!!!");
				return -1;
			} finally {
				rwLock.readLock().unlock();
			}
		}

		@Override
		public void run() {
			int number = get();
			if (number >= 0) {
				System.out.println("\tReader " + Thread.currentThread().threadId() + " -> get: " + number);
			}
		}
	}
}
