package l10_concurrent;

import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.locks.ReentrantLock;

/*
	Java Lock API provides more visibility and options for locking, unlike
	synchronized where a thread might end up waiting indefinitely for the lock,
	we can use tryLock() to make sure thread waits for specific time only.

	Synchronization code is much cleaner and easy to maintain whereas with Lock
	we are forced to have try-finally block to make sure Lock is released even
	if some exception is thrown between lock() and unlock() method calls.
	Synchronization blocks or methods can cover only one method whereas we can
	acquire the lock in one method and release it in another method with Lock
	API.

	"synchronized" keyword doesn't provide fairness whereas we can set fairness
	to true while creating ReentrantLock object so that longest waiting thread
	gets the lock first.

	We can create different conditions for Lock and different thread can await()
	for different conditions.
*/

public class ReentrantLockExample {
	private int countBasic = 0;
	private int countLocked = 0;
	private final ReentrantLock lock = new ReentrantLock();

	void incrementNoLock() {
		++countBasic;
	}

	void incrementLock() {
		lock.lock();
		try {
			++countLocked;
		} finally { // to make sure that unlock() is called
			lock.unlock();
		}
	}

	interface Increment {
		void increment();
	}

	void incrementLoop(Increment increment) {
		try (ExecutorService executor = Executors.newFixedThreadPool(8)) {
			for (int i = 0; i < 1000000; ++i) {
				executor.submit(new Runnable() {
					@Override
					public void run() {
						increment.increment();
					}
				});
			}
		}
	}

	void runTest() {
		incrementLoop(this::incrementNoLock);
		incrementLoop(this::incrementLock);

		System.out.println("countBasic : " + countBasic);
		System.out.println("countLocked: " + countLocked);
	}

	public static void main(String[] args) {
		new ReentrantLockExample().runTest();
	}
}
